#!/usr/bin/env python3
"""Start the LNXlink service"""

import os
import sys
import time
import json
import threading
import logging
import argparse
import platform
import traceback

import distro
from lnxlink import modules
from lnxlink import config_setup
from lnxlink import files_setup
from lnxlink.mqtt import MQTT
from lnxlink.system_monitor import MonitorSuspend, GracefulKiller
from lnxlink.modules.scripts import helpers

version, path = files_setup.get_version()
logger = logging.getLogger("lnxlink")


# pylint: disable=too-many-instance-attributes
class LNXlink:
    """Start LNXlink service that loads all modules and connects to MQTT"""

    version = version
    path = path

    def __init__(self, config):
        logger.info("LNXlink %s, Python %s", self.version, platform.python_version())
        logger.debug("Path=%s", self.path)
        self.config = config
        self.config_path = config["config_path"]
        self.kill = None
        self.display = None
        self.inference_times = {}
        self.addons = {}
        self.prev_publish = {}
        self.saved_publish = {}
        self.update_change_interval = 900

        # Read configuration from yaml file
        self.publ_queue = files_setup.UniqueQueue()
        self.mqtt = MQTT(self.config)
        self.stop_event = threading.Event()

    def start(self, exclude_modules_arg):
        """Run each addon included in the modules folder"""
        conf_exclude = self.config["exclude"]
        conf_exclude = [] if conf_exclude is None else conf_exclude
        conf_exclude.extend(exclude_modules_arg)
        loaded_modules = modules.parse_modules(
            self.config["modules"], self.config["custom_modules"], conf_exclude
        )
        for _, addon in loaded_modules.items():
            try:
                tmp_addon = addon(self)
                self.addons[addon.service] = tmp_addon
            except Exception as err:
                logger.error(
                    "Error with addon %s, please remove it from your config: %s",
                    addon.service,
                    err,
                )
                logger.debug(
                    traceback.format_exc(),
                )
        loaded = list(self.addons.keys())
        loaded.sort()
        logger.info("Loaded addons: %s", ", ".join(loaded))

        mqtt_status = self.mqtt.setup_mqtt(self.on_connect, self.on_message)
        threading.Thread(target=self.monitor_run, daemon=True).start()
        threading.Thread(target=self.monitor_queue, daemon=True).start()
        return mqtt_status

    def add_settings(self, name, settings):
        """Adds missing configuration under settings"""
        self.config = config_setup.add_settings(self.config, name, settings)

    def publish_monitor_data(self, name, pub_data, retain=True):
        """Publish info data to mqtt in the correct format"""
        subtopic = helpers.text_to_topic(name)
        topic = f"{self.config['pref_topic']}/monitor_controls/{subtopic}"
        if pub_data is None:
            return
        if isinstance(pub_data, bool):
            if pub_data is True:
                pub_data = "ON"
            if pub_data is False:
                pub_data = "OFF"
        if isinstance(pub_data, dict):
            if all(v is None for v in pub_data.values()):
                return
        if isinstance(pub_data, list):
            if all(v is None for v in pub_data):
                return
        if isinstance(pub_data, (dict, list)):
            pub_data = json.dumps(pub_data)

        # User option that checks and sends only the updated sensors
        update_change_time = time.time() - self.prev_publish.get("last_update", 0)
        if update_change_time > self.update_change_interval:
            self.prev_publish = {"last_update": time.time()}
        if (
            self.config["update_on_change"] or isinstance(pub_data, bytes)
        ) and self.prev_publish.get(topic) == pub_data:
            return

        self.prev_publish[topic] = pub_data
        self.saved_publish[subtopic.replace("/", "_")] = pub_data
        self.mqtt.publish(topic, pub_data, retain)

    def run_module(self, name, method, retain=True):
        """Runs the method of a module"""
        try:
            start_time = time.time()
            if isinstance(method, (dict, list, bool, bytes, int, str, float)):
                pub_data = method
            else:
                pub_data = method()
                diff_time = round(time.time() - start_time, 5)
                self.inference_times[name] = diff_time
            self.publ_queue.add_item(name, pub_data, retain)
        except Exception as err:
            logger.error(
                "Error with addon %s: %s, %s",
                name,
                err,
                traceback.format_exc(),
            )

    def monitor_run(self):
        """Gets information from each Addon and adds it to the queue"""
        while not self.stop_event.is_set():
            methods_to_run = []
            if not self.kill:
                for _, addon in self.addons.items():
                    if hasattr(addon, "get_info"):
                        methods_to_run.append(
                            {
                                "name": addon.name,
                                "method": addon.get_info,
                            }
                        )
                self.mqtt.send_lwt("ON")
                for method in methods_to_run:
                    self.run_module(method["name"], method["method"])
            if self.stop_event.wait(timeout=self.config["update_interval"]):
                break
        logger.info("Stopped monitor_run")

    def monitor_queue(self):
        """Loop through the queue list and publish data to MQTT broker"""
        while not self.stop_event.is_set():
            if not self.kill:
                time.sleep(0.01)
                for name, queue_data in self.publ_queue:
                    pub_data, retain = queue_data
                    self.publish_monitor_data(name, pub_data, retain)
                    time.sleep(0.01)
            if self.stop_event.wait(timeout=0.2):
                self.mqtt.send_lwt("OFF")
                break
        logger.info("Stopped monitor_queue")

    def on_connect(self, client, userdata, flags, rcode, *args):
        """Callback for MQTT connect which reports the connection status
        back to MQTT server"""
        logger.info("MQTT connection: %s", self.mqtt.get_rcode_name(rcode))
        client.subscribe(f"{self.config['pref_topic']}/commands/#")
        self.mqtt.send_lwt("ON")
        if self.config["mqtt"]["discovery"]["enabled"]:
            self.setup_discovery()
        if self.kill is None:
            self.kill = False

    def disconnect(self, *args):
        """Service has stopped"""
        self.kill = True
        self.mqtt.disconnect()
        self.stop_event.set()

    def replace_values_with_none(self, data):
        """Replaces specified values with None recursively"""
        if isinstance(data, (str, bool, float, int)):
            return None
        if isinstance(data, dict):
            return {
                key: self.replace_values_with_none(value) for key, value in data.items()
            }
        return data

    def temp_connection_callback(self, status):
        """Report the connection status to MQTT server"""
        logger.info("Got Temp Connection Callback with status %s", status)
        self.kill = True
        if status:
            logger.info("Power Down detected.")
            self.mqtt.send_lwt("OFF")
            if self.config["mqtt"]["clear_on_off"]:
                for topic, message in self.prev_publish.items():
                    message = self.replace_values_with_none(message)
                    self.mqtt.publish(topic, message)
        else:
            logger.info("Power Up detected.")
            if self.kill:
                self.kill = False
            self.mqtt.send_lwt("ON")

    def on_message(self, client, userdata, msg):
        """MQTT message is received with a module command to execute"""
        topic = msg.topic.replace(f"{self.config['pref_topic']}/commands/", "")
        message = msg.payload
        logger.info("Message received %s: %s", topic, message)
        try:
            message = json.loads(message)
        except json.decoder.JSONDecodeError:
            message = message.decode()
        except Exception as err:
            logger.debug("Error reading message: %s", err)

        service = topic.split("/")
        addon = self.addons.get(service[0])
        if addon is not None:
            if hasattr(addon, "start_control"):
                threading.Thread(
                    target=self.start_control_bg,
                    args=(addon, topic, service, message),
                    daemon=True,
                ).start()

    def start_control_bg(self, addon, topic, service, message):
        """Starts the start_control method of a module in the background"""
        try:
            result = addon.start_control(service, message)
            if result is not None:
                result_topic = (
                    f"{self.config['pref_topic']}/command_result/{topic.strip('/')}"
                )
                self.mqtt.publish(result_topic, result)
        except Exception as err:
            logger.error(
                "Couldn't run command for module %s: %s, %s",
                addon,
                err,
                traceback.format_exc(),
            )

    # pylint: disable=too-many-locals
    def setup_discovery_entities(self, addon, service, exp_name, options):
        """Send discovery information on Home Assistant for controls"""
        discovery_template = {
            "availability": {
                "topic": f"{self.config['pref_topic']}/lwt",
                "payload_available": "ON",
                "payload_not_available": "OFF",
            },
            "device": {
                "identifiers": [self.config["mqtt"]["clientId"]],
                "name": self.config["mqtt"]["clientId"],
                "model": f"{distro.name()} {distro.version()}",
                "manufacturer": "LNXlink",
                "sw_version": version,
            },
        }
        subtopic = helpers.text_to_topic(addon.name)
        if "method" in options or options.get("subtopic", False):
            subcontrol = helpers.text_to_topic(exp_name)
            subtopic = f"{subtopic}/{subcontrol}"
        state_topic = f"{self.config['pref_topic']}/monitor_controls/{subtopic}"
        control_name_topic = helpers.text_to_topic(exp_name)
        command_topic = (
            f"{self.config['pref_topic']}/commands/{service}/{control_name_topic}"
        )

        lookup_options = {
            "value_template": {
                "value_template": options.get("value_template", ""),
            },
            "attributes_template": {
                "json_attributes_topic": state_topic,
                "json_attributes_template": options.get(
                    "attributes_template", "{{ value_json | tojson }}"
                ),
            },
            "icon": {"icon": options.get("icon", "")},
            "unit": {"unit_of_measurement": options.get("unit", "")},
            "title": {"title": options.get("title", "")},
            "entity_picture": {"entity_picture": options.get("entity_picture", "")},
            "device_class": {"device_class": options.get("device_class", "")},
            "state_class": {"state_class": options.get("state_class", "")},
            "entity_category": {"entity_category": options.get("entity_category", "")},
            "enabled": {"enabled_by_default": options.get("enabled", True)},
            "expire_after": {"expire_after": options.get("expire_after", "")},
            "install": {
                "command_topic": command_topic,
                "payload_install": options.get("install", ""),
            },
        }
        lookup_entities = {
            "sensor": {
                "state_topic": state_topic,
            },
            "binary_sensor": {
                "state_topic": state_topic,
            },
            "camera": {
                "topic": state_topic,
                "image_encoding": options.get("encoding"),
            },
            "image": {
                "image_topic": state_topic,
                "image_encoding": options.get("encoding"),
            },
            "update": {"state_topic": state_topic},
            "button": {
                "command_topic": command_topic,
                "payload_press": options.get("payload_press", "PRESS"),
            },
            "switch": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "payload_off": options.get("command_off", "OFF"),
                "payload_on": options.get("command_on", "ON"),
                "state_off": "OFF",
                "state_on": "ON",
            },
            "text": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "min": options.get("min", 0),
                "max": options.get("max", 255),
            },
            "number": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "min": options.get("min", 1),
                "max": options.get("max", 100),
                "step": options.get("step", 1),
            },
            "select": {
                "state_topic": state_topic,
                "command_topic": command_topic,
                "options": options.get("options", []),
            },
            "device_tracker": {
                "json_attributes_topic": state_topic,
            },
            "media_player": {
                "name": self.config["mqtt"]["clientId"],
                "state_state_topic": f"{state_topic}/state",
                "state_title_topic": f"{state_topic}/title",
                "state_artist_topic": f"{state_topic}/artist",
                "state_album_topic": f"{state_topic}/album",
                "state_duration_topic": f"{state_topic}/duration",
                "state_position_topic": f"{state_topic}/position",
                "state_volume_topic": f"{state_topic}/volume",
                "state_albumart_topic": f"{state_topic}/albumart",
                "state_mediatype_topic": f"{state_topic}/mediatype",
                "command_volume_topic": f"{command_topic}/set_volume",
                "command_play_topic": f"{command_topic}/play",
                "command_play_payload": "Play",
                "command_pause_topic": f"{command_topic}/pause",
                "command_pause_payload": "Pause",
                "command_playpause_topic": f"{command_topic}/playpause",
                "command_playpause_payload": "PlayPause",
                "command_next_topic": f"{command_topic}/next",
                "command_next_payload": "Next",
                "command_previous_topic": f"{command_topic}/previous",
                "command_previous_payload": "Previous",
                "command_playmedia_topic": f"{command_topic}/play_media",
            },
            "notify": {
                "command_topic": command_topic,
                "json_attributes_topic": state_topic,
                "name": self.config["mqtt"]["clientId"],
            },
        }
        discovery = discovery_template.copy()
        discovery["name"] = exp_name
        discovery[
            "unique_id"
        ] = f"{self.config['mqtt']['clientId']}_{control_name_topic}"
        discovery.update(lookup_entities.get(options["type"], {}))
        for option in options:
            discovery.update(lookup_options.get(option, {}))

        if options["type"] not in lookup_entities:
            logger.error("Not supported: %s", options["type"])
            return
        if "value_template" in discovery and options["type"] in ["camera", "image"]:
            del discovery["json_attributes_topic"]
            del discovery["json_attributes_template"]
        discovery_prefix = self.config["mqtt"]["discovery"]["prefix"]
        self.mqtt.publish(
            f"{discovery_prefix}/{options['type']}/lnxlink/{discovery['unique_id']}/config",
            payload=json.dumps(discovery),
        )
        if options["type"] == "media_player":
            logger.info(
                "MQTT Media Player configuration name: lnxlink/%s",
                discovery["unique_id"],
            )

    def setup_discovery(self, filter_name=None):
        """First time setup of discovery for Home Assistant"""
        for service, addon in self.addons.items():
            if filter_name is not None and filter_name != service:
                continue
            if hasattr(addon, "exposed_controls"):
                for exp_name, options in addon.exposed_controls().items():
                    try:
                        self.setup_discovery_entities(addon, service, exp_name, options)
                    except Exception as err:
                        logger.error(
                            "%s: %s, %s", exp_name, err, traceback.format_exc()
                        )

    def restart_script(self):
        """Restarts itself"""
        logger.info("Restarting LNXlink...")
        os.execv(sys.executable, ["python"] + sys.argv)


def main():
    """Starts the app with some arguments"""
    description = (
        f"LNXlink {version} bridges this OS with Home Assistant through an MQTT broker."
    )
    parser = argparse.ArgumentParser(prog="lnxlink", description=description)
    parser.add_argument(
        "-c",
        "--config",
        help="Configuration file",
    )
    parser.add_argument(
        "-i",
        "--ignore-systemd",
        help="Runs without setting up SystemD service",
        action="store_true",
    )
    parser.add_argument(
        "-e",
        "--exclude",
        help="Exclude modules from running",
        default=[],
        type=lambda t: [s.strip() for s in t.split(",")],
    )
    parser.add_argument(
        "-l",
        "--logging",
        help="Set the logging level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "-s",
        "--setup",
        help="Runs only the setup configuration workflow",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Shows current version of LNXlink",
        action="store_true",
    )
    args = parser.parse_args()

    if args.version:
        print(version)
        sys.exit()
    if args.config is None:
        parser.print_help()
        parser.exit("\nSomething went wrong, --config condition was not set")
    config_path = os.path.abspath(args.config)
    files_setup.setup_logger(config_path, args.logging)
    config_setup.setup_config(config_path)
    if args.setup:
        logger.info("The configuration exists under the file: %s", config_path)
        sys.exit()
    if not args.ignore_systemd:
        config_setup.setup_systemd(config_path)
    else:
        logger.info(
            "By not setting up the SystemD, LNXlink won't be able to start on boot..."
        )

    config = files_setup.read_config(config_path)
    lnxlink = LNXlink(config)

    # Monitor for system changes (Shutdown/Suspend/Sleep)
    monitor_suspend = MonitorSuspend(lnxlink.temp_connection_callback)
    monitor_suspend.start()
    monitor_gracefulkiller = GracefulKiller(lnxlink.temp_connection_callback)

    # Starts the main app
    start_status = lnxlink.start(args.exclude)
    if start_status:
        while not monitor_gracefulkiller.kill_now:
            time.sleep(1.0)
        monitor_suspend.stop()
        lnxlink.disconnect()
    else:
        monitor_suspend.stop()


if __name__ == "__main__":
    main()



"""Information and control of pulseaudio devices"""
import json
import logging
from lnxlink.modules.scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Audio Select"
        self.lnxlink = lnxlink
        self.audio_system = self._get_audio_system()
        if self.audio_system is None:
            raise SystemError("Can't find any supported audio system")
        self.devices = {
            "speaker": {},
            "microphone": {},
            "defaults": {"microphone": "", "speaker": ""},
            "changed": False,
        }
        self._get_devices()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {
            "Microphone Select": {
                "type": "select",
                "icon": "mdi:microphone-variant",
                "options": list(self.devices["microphone"].keys()),
                "value_template": "{{ value_json.microphone }}",
            },
            "Speaker Select": {
                "type": "select",
                "icon": "mdi:audio-input-rca",
                "options": list(self.devices["speaker"].keys()),
                "value_template": "{{ value_json.speaker }}",
            },
        }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        self._get_devices()
        if self.devices["changed"]:
            self.lnxlink.setup_discovery("audio_select")
        return self.devices["defaults"]

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "microphone_select":
            if self.audio_system == "pactl":
                syscommand(
                    f"pactl set-default-source {self.devices['microphone'][data]}"
                )
            elif self.audio_system == "pulsectl":
                with self.pulsectl.Pulse("lnxlink_ctl") as pulse:
                    pulse.default_set(
                        pulse.get_source_by_name(self.devices["microphone"][data])
                    )
        if topic[1] == "speaker_select":
            if self.audio_system == "pactl":
                syscommand(f"pactl set-default-sink {self.devices['speaker'][data]}")
            elif self.audio_system == "pulsectl":
                with self.pulsectl.Pulse("lnxlink_ctl") as pulse:
                    pulse.default_set(
                        pulse.get_sink_by_name(self.devices["speaker"][data])
                    )

    def _get_audio_system(self):
        """Get system volume type"""
        _, _, returncode = syscommand(
            "pactl get-sink-volume @DEFAULT_SINK@", ignore_errors=True
        )
        if returncode == 0:
            return "pactl"

        self.pulsectl = import_install_package("pulsectl", ">=23.5.2")
        if self.pulsectl is not None:
            return "pulsectl"

        return None

    # pylint: disable=too-many-branches
    def _get_devices(self):
        """Get a list of all audio devices"""
        devices = {
            "speaker": {},
            "microphone": {},
            "defaults": {"microphone": "", "speaker": ""},
            "changed": False,
        }

        if self.audio_system == "pactl":
            stdout, _, _ = syscommand("pactl -f json list sinks")
            for sink in json.loads(stdout):
                devices["speaker"][sink["description"]] = sink["name"]
            stdout, _, _ = syscommand("pactl -f json list sources")
            for source in json.loads(stdout):
                devices["microphone"][source["description"]] = source["name"]
            stdout, _, _ = syscommand("pactl get-default-sink")
            for description, name in devices["speaker"].items():
                if name == stdout:
                    devices["defaults"]["speaker"] = description
            stdout, _, _ = syscommand("pactl get-default-source")
            for description, name in devices["microphone"].items():
                if name == stdout:
                    devices["defaults"]["microphone"] = description
        elif self.audio_system == "pulsectl":
            try:
                with self.pulsectl.Pulse("lnxlink") as pulse:
                    for sink in pulse.sink_list():
                        devices["speaker"][sink.description] = sink.name
                    for source in pulse.source_list():
                        devices["microphone"][source.description] = source.name

                    server_info = pulse.server_info()
                    for description, name in devices["speaker"].items():
                        if name == server_info.default_sink_name:
                            devices["defaults"]["speaker"] = description
                    for description, name in devices["microphone"].items():
                        if name == server_info.default_source_name:
                            devices["defaults"]["microphone"] = description
            except Exception as err:
                logger.error("Error with Pulseaudio: %s", err)

        if len(devices["defaults"]["speaker"]) != len(
            self.devices["defaults"].get("speaker", [])
        ):
            devices["changed"] = True
        if len(devices["defaults"]["microphone"]) != len(
            self.devices["defaults"].get("microphone", [])
        ):
            devices["changed"] = True
        self.devices = devices

        return devices



"""Controls mouse"""
import os
import time
import logging
import threading
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand, get_display_variable

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Mouse"
        self.movement = [0, 0]
        commands = {
            "ydotool": {
                "left_click": "ydotool click 0xc0",
                "right_click": "ydotool click 0xc1",
                "left_mouse_down": "ydotool click 0x40",
                "left_mouse_up": "ydotool click 0x80",
                "move_rel": "ydotool mousemove -x %s -y %s",
                "move_abs": "ydotool mousemove -a -x %s -y %s",
                "run_check": "ydotool mousemove",
            },
            "xdotool": {
                "left_click": "xdotool click 1",
                "right_click": "xdotool click 3",
                "left_mouse_down": "ydotool mousedown 1",
                "left_mouse_up": "ydotool mouseup 1",
                "move_rel": "xdotool mousemove -- %s %s",
                "move_abs": "xdotool mousemove %s %s",
            },
        }
        self.commands = None
        for command, command_options in commands.items():
            if which(command) is not None:
                run_check = command_options.get("run_check")
                if run_check is not None:
                    _, _, returncode = syscommand(run_check)
                    if returncode != 0:
                        continue
                self.commands = command_options
                logger.debug("Using '%s' for mouse control", command)
                break
        if self.commands is None:
            raise SystemError("System commands 'ydotool' or 'xdotool' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Mouse Coordinates": {
                "type": "text",
                "icon": "mdi:mouse-variant",
            },
            "Mouse Left": {
                "type": "button",
                "icon": "mdi:gamepad-circle-left",
            },
            "Mouse Right": {
                "type": "button",
                "icon": "mdi:gamepad-circle-right",
            },
            "Mouse Up": {
                "type": "button",
                "icon": "mdi:gamepad-circle-up",
            },
            "Mouse Down": {
                "type": "button",
                "icon": "mdi:gamepad-circle-down",
            },
            "Mouse Click": {
                "type": "button",
                "icon": "mdi:mouse-left-click",
            },
            "Mouse Click Right": {
                "type": "button",
                "icon": "mdi:mouse-right-click",
            },
            "Mouse Click Left Up": {
                "type": "button",
                "icon": "mdi:mouse-move-up",
            },
            "Mouse Click Left Down": {
                "type": "button",
                "icon": "mdi:mouse-move-down",
            },
        }

    # pylint: disable=too-many-branches
    def start_control(self, topic, data):
        """Control system"""
        display_variable = get_display_variable()
        print(display_variable)
        if display_variable is not None and os.environ["DISPLAY"] is None:
            os.environ["DISPLAY"] = display_variable
            logger.info("Initializing empty DISPLAY environment variable")

        if topic[1] == "mouse_coordinates":
            if "," in data:
                coords = data.replace(" ", "").split(",")
            elif " " in data:
                coords = data.split(" ")
            else:
                return
            move_type = "move_abs"
            if coords[0].startswith("+") or coords[0].startswith("-"):
                move_type = "move_rel"
            if coords[1].startswith("+") or coords[1].startswith("-"):
                move_type = "move_rel"
            syscommand(self.commands[move_type] % (coords[0], coords[1]))
        elif topic[1] == "mouse_left":
            self._move([-1, 0])
        elif topic[1] == "mouse_right":
            self._move([1, 0])
        elif topic[1] == "mouse_up":
            self._move([0, -1])
        elif topic[1] == "mouse_down":
            self._move([0, 1])
        elif topic[1] in ["mouse_click", "mouse_click_left"]:
            self.movement = [0, 0]
            syscommand(self.commands["left_click"])
        elif topic[1] == "mouse_click_right":
            self.movement = [0, 0]
            syscommand(self.commands["right_click"])
        elif topic[1] == "mouse_click_left_down":
            syscommand(self.commands["left_mouse_down"])
        elif topic[1] == "mouse_click_left_up":
            self.movement = [0, 0]
            syscommand(self.commands["left_mouse_up"])

    def _move(self, movement):
        if self.movement == movement:
            # Stops mouse if it is moving
            self.movement = [0, 0]
        else:
            self.movement = movement
            move_mouse_thr = threading.Thread(target=self._move_mouse)
            move_mouse_thr.start()

    def _move_mouse(self):
        movement = self.movement
        for i in range(1, 90):
            if self.movement != movement:
                # Stops mouse if a new move command is sent
                return
            move_x = movement[0] * i
            move_y = movement[1] * i
            syscommand(self.commands["move_rel"] % (move_x, move_y))
            time.sleep(0.05)
        # Used if the user wants to run the same movement
        self.movement = [0, 0]



"""WOL Status control"""
import re
from shutil import which
import psutil
from lnxlink.modules.scripts.helpers import syscommand, text_to_topic


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "WOL"
        self.lnxlink = lnxlink
        self.interfaces = {}
        if which("ethtool") is None:
            raise SystemError("System command 'ethtool' not found")
        self._get_interfaces()
        if len(self.interfaces) == 0:
            raise SystemError(
                "Can't find any supported interface or ethtool doesn't has access to sudo."
            )

    def get_info(self):
        """Gather information from the system"""
        for interf in self.interfaces:
            cmd = f"sudo -n ethtool {interf}"
            stdout, _, _ = syscommand(cmd)
            self.interfaces[interf] = "OFF"
            match_wol = re.search(r"\tWake-on: (\S*)", stdout)
            if match_wol:
                if "g" in match_wol.group(1):
                    self.interfaces[interf] = "ON"

        return self.interfaces

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for interf in self.interfaces:
            discovery_info[f"WOL {interf}"] = {
                "type": "switch",
                "icon": "mdi:ethernet",
                "value_template": f"{{{{ value_json.get('{interf}') }}}}",
            }
        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        for interf in self.interfaces:
            if text_to_topic(topic[1]).replace("wol_", "") == interf:
                if data == "ON":
                    syscommand(f"sudo ethtool -s {interf} wol g")
                elif data == "OFF":
                    syscommand(f"sudo ethtool -s {interf} wol d")

    def _get_interfaces(self):
        """Get a list of all interfaces"""
        interfaces = []
        addrs = psutil.net_if_addrs()
        for interf, _ in addrs.items():
            if interf.startswith("veth"):
                continue
            interfaces.append(interf)

        self.interfaces = {}
        for interf in interfaces:
            cmd = f"sudo -n ethtool {interf}"
            stdout, _, errorcode = syscommand(cmd)
            if errorcode == 0:
                match = re.search(r"\tSupports Wake-on: (\S*)", stdout)
                if match:
                    if "g" in match.group(1):
                        self.interfaces[interf] = None


if __name__ == "__main__":
    Addon("").get_info()



"""Monitors for shutdown/sleep events"""

import time
import threading
import signal
import logging
import traceback
from jeepney import MatchRule, DBus
from jeepney.io.blocking import open_dbus_connection, Proxy


logger = logging.getLogger("lnxlink")


class MonitorSuspend:
    """Monitor DBUS for Suspend and Shutdown events"""

    def __init__(self, callback):
        self.callback = callback
        self.use = False
        try:
            self.connection = open_dbus_connection(bus="SYSTEM")
            bus_proxy = Proxy(DBus(), self.connection)
            sleep_match = MatchRule(
                type="signal",
                interface="org.freedesktop.login1.Manager",
                member="PrepareForSleep",
                path="/org/freedesktop/login1",
            )
            bus_proxy.AddMatch(sleep_match)

            self.timer1 = threading.Thread(target=self.watch_loop, daemon=True)
            self.use = True
        except Exception as err:
            logger.error(
                "Can't use DBus: %s, %s",
                err,
                traceback.format_exc(),
            )

    def watch_loop(self):
        """Run the dbus check for new sleep messages"""
        errors = 0
        while True:
            try:
                message = self.connection.receive()
                self.callback(message.body[0])
                errors = 0
            except Exception as err:
                logger.error(
                    "DBus Error: %s, %s",
                    err,
                    traceback.format_exc(),
                )
                time.sleep(1)
                errors += 1
                if errors > 10:
                    break

    def start(self):
        """Start the timer of the thread"""
        if self.use:
            try:
                self.timer1.start()
            except Exception as err:
                logger.error(
                    "Can't start systemMonitor: %s, %s",
                    err,
                    traceback.format_exc(),
                )

    def stop(self):
        """Stop the timer of the thread"""
        if self.use:
            try:
                self.connection.close()
            except Exception as err:
                logger.error(
                    "Can't stop systemMonitor: %s, %s",
                    err,
                    traceback.format_exc(),
                )


class GracefulKiller:
    """Monitor Signal Termination"""

    def __init__(self, callback):
        """Checks for termination signals"""
        self.kill_now = False
        self.callback = callback
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum, frame):
        """Stop the app gracefully"""
        logger.info("stopped gracefully")
        self.callback(1)
        self.kill_now = True



"""List all steam games and start them when selected"""
import os
import re
import glob
import struct
import binascii
import logging
from shutil import which
import psutil
from lnxlink.modules.scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Steam"
        self.lnxlink = lnxlink
        if which("steam") is None:
            raise SystemError("Steam not found")
        self._requirements()
        self.steam_config = self._find_libary_config()
        if self.steam_config is None:
            raise SystemError("Steam library can't be found")
        self.lnxlink.add_settings("steam", {"library": self.steam_config})
        self.games = {}
        self.games = self._get_games()

    def _requirements(self):
        self.vdf = import_install_package("vdf", ">=3.4")

    def exposed_controls(self):
        """Exposes to home assistant"""
        mygames = list(self.games.values())
        mygames.sort()
        mygames.insert(0, "Stopped")
        discovery_info = {
            "Steam Game Select": {
                "type": "select",
                "icon": "mdi:steam",
                "options": mygames,
            },
        }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        games = self._get_games()
        if self.games != games:
            self.lnxlink.setup_discovery("steam")
        self.games = games
        return self._get_current_game()

    def start_control(self, topic, data):
        """Control system"""
        try:
            position = list(self.games.values()).index(data)
            game_id = list(self.games.keys())[position]
            logger.info("running command: steam steam://rungameid/%s", game_id)
            syscommand(f"steam steam://rungameid/{game_id}")
        except ValueError:
            logger.error("Can't find selected game: %s", data)

    def _get_current_game(self):
        """Checks running processes"""
        app_ids = set()
        for p in psutil.process_iter():
            try:
                process = " ".join(p.cmdline())
            except psutil.ZombieProcess:
                process = ""
            except psutil.NoSuchProcess:
                process = ""
            if "AppId=" in process:
                match = re.findall(r"AppId=(\d+) ", process)
                if match:
                    app_ids.add(match[0])
        if len(app_ids) > 1:
            logger.info(
                "More than one game instance has been found: %s", ",".join(app_ids)
            )
        for app_id in app_ids:
            game = self.games.get(app_id)
            if game is None:
                long_app_id = int(app_id) << 32 | 0x02000000
                game = self.games.get(long_app_id)
            if game is not None:
                return game
            logger.error("Can't find game app id: %s", app_id)

        return "Stopped"

    def _find_libary_config(self, level=7):
        """Finds the library folders configuration from home directory
        with a max depth level"""
        settings_library = (
            self.lnxlink.config["settings"].get("steam", {}).get("library")
        )
        if settings_library is not None and os.path.exists(settings_library):
            return settings_library
        home_dir = os.path.expanduser("~")
        num_sep = home_dir.count("/")
        name = "libraryfolders.vdf"
        for root, dirs, files in os.walk(home_dir):
            num_sep_this = root.count("/")
            if num_sep + level <= num_sep_this:
                del dirs[:]
            if name in files:
                return os.path.join(root, name)
        return None

    # pylint: disable=too-many-locals
    def _get_games(self):
        """Gets a dictionary of the game_ids and their names"""
        with open(self.steam_config, encoding="UTF-8") as file:
            libraries = self.vdf.load(file)

        # Steam Games
        games = {}
        for num in libraries["libraryfolders"].values():
            for app_id in num["apps"].keys():
                if app_id not in self.games:
                    acf_path = os.path.join(
                        num["path"], f"steamapps/appmanifest_{app_id}.acf"
                    )
                    if os.path.exists(acf_path):
                        with open(acf_path, encoding="UTF-8") as file:
                            game_properties = self.vdf.load(file)
                        games[app_id] = game_properties["AppState"]["name"]
                else:
                    games[app_id] = self.games[app_id]

        # Non-Steam Games
        shortcut_dirs = os.path.abspath(
            f"{self.steam_config}/../../userdata/*/config/shortcuts.vdf"
        )
        for shortcut_dir in glob.glob(shortcut_dirs):
            with open(shortcut_dir, "rb") as file:
                shortcuts_dict = self.vdf.binary_loads(file.read())
            for shortcut in shortcuts_dict["shortcuts"].values():
                # Convert to HEX
                bin_appid = struct.Struct("<i").pack(shortcut["appid"])
                hex_appid = binascii.hexlify(bin_appid).decode("UTF-8")

                # Convert to long_appid
                reversed_hex = bytes.fromhex(hex_appid)[::-1].hex()
                long_app_id = int(reversed_hex, 16) << 32 | 0x02000000
                games[long_app_id] = shortcut["AppName"]

        return games



"""Gets the active window"""
# Wayland requires either of the extensions
# - [flexagoon/focused-window-dbus](https://github.com/flexagoon/focused-window-dbus)
# - [bkbilly/GnomeShell-WindowQueryTool](https://github.com/bkbilly/GnomeShell-WindowQueryTool)
import os
import subprocess
import json
import logging
from lnxlink.modules.scripts.helpers import import_install_package, get_display_variable


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Active Window"
        self._requirements()

    def _requirements(self):
        self.lib = {
            "ewmh": import_install_package("ewmh", ">=0.1.6"),
            "xlib": import_install_package("python-xlib", ">=0.33", "Xlib.display"),
        }

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Active Window": {
                "type": "sensor",
                "icon": "mdi:book-open-page-variant",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        display_variable = get_display_variable()
        # Detect session type
        session_type = os.environ.get("XDG_SESSION_TYPE", "unknown").lower()

        logger = logging.getLogger("lnxlink")
        logger.debug("Session type: %s", session_type)

        if "x11" == session_type:
            # X11
            if display_variable is None:
                return ""
            display = self.lib["xlib"].display.Display(display_variable)
            ewmh = self.lib["ewmh"].EWMH(_display=display)
            win = ewmh.getActiveWindow()
            window_name = ewmh.getWmName(win)

            if window_name is None:
                return None
            return window_name.decode()

        if "wayland" == session_type:
            window_name = None

            # Wayland: GNOME WindowQueryTool extension
            try:
                method = "org.gnome.Shell.Extensions.WindowQueryTool.GetWindowInfo"
                result = subprocess.run(
                    [
                        "gdbus",
                        "call",
                        "--session",
                        "--dest",
                        "org.gnome.Shell",
                        "--object-path",
                        "/org/gnome/Shell/Extensions/WindowQueryTool",
                        "--method",
                        method,
                    ],
                    capture_output=True,
                    text=True,
                    timeout=5,
                    check=True,
                )

                if result.returncode == 0 and result.stdout.strip():
                    json_str = result.stdout.strip("()',\n")
                    data = json.loads(json_str)
                    window_name = data.get("focused_window_title", None)
                    return window_name
                logger.debug("Method not found: %s", method)
            except (subprocess.TimeoutExpired, FileNotFoundError, json.JSONDecodeError):
                pass

            # Wayland: GNOME FocusedWindow extension
            try:
                method = "org.gnome.shell.extensions.FocusedWindow.Get"
                result = subprocess.run(
                    [
                        "gdbus",
                        "call",
                        "--session",
                        "--dest",
                        "org.gnome.Shell",
                        "--object-path",
                        "/org/gnome/shell/extensions/FocusedWindow",
                        "--method",
                        method,
                    ],
                    capture_output=True,
                    text=True,
                    timeout=5,
                    check=True,
                )

                if result.returncode == 0 and result.stdout.strip():
                    json_str = result.stdout.strip("()',\n")
                    data = json.loads(json_str)
                    window_name = data.get("title", None)
                    return window_name
                logger.debug("Method not found: %s", method)
            except (subprocess.TimeoutExpired, FileNotFoundError, json.JSONDecodeError):
                pass

        return None



"""Gets the coordinates of the maching using WiFi Networks"""
import io
import csv
import time
import logging
from shutil import which

import requests
from lnxlink.modules.scripts.helpers import syscommand


logger = logging.getLogger("lnxlink")


def scan_wifi_nmcli():
    """Command to list Wi-Fi access points with their details"""
    command = ["nmcli", "-t", "-f", "BSSID,SSID,SIGNAL", "dev", "wifi", "list"]
    stdout, _, _ = syscommand(command, timeout=7)
    output_buffer = io.StringIO(stdout)
    csv_reader = csv.reader(output_buffer, delimiter=":", escapechar="\\")

    wifi_aps = []
    for row in csv_reader:
        if len(row) >= 3:
            bssid = row[0].strip().upper()
            ssid = row[1].strip()
            signal_str = row[2].strip()

            if ssid == "--" or ssid.lower() == "null":
                ssid = None  # Represent hidden/empty SSIDs as None
            try:
                signal = int(signal_str)
            except ValueError:
                continue

            wifi_aps.append(
                {
                    "macAddress": bssid,
                    "ssid": ssid,
                    "signalStrength": signal,
                }
            )
    return wifi_aps


def get_location_from_beacondb(wifi_aps_data, consider_ip=False):
    """Get the coordinates from BeaconDB"""
    try:
        response = requests.post(
            "https://api.beacondb.net/v1/geolocate",
            headers={
                "Content-Type": "application/json",
            },
            json={
                "wifiAccessPoints": wifi_aps_data,
                "considerIp": consider_ip,
            },
        )
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as err:
        logger.error("Error querying BeaconDB: %s", err)
        return None


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "BeaconDB"
        self.lnxlink = lnxlink
        self.last_time = 0
        self.update_interval = 360  # Check for position every 6 minutes
        self.position = None
        if which("nmcli") is None:
            raise SystemError("System command 'nmcli' not found")
        self.lnxlink.add_settings("beacondb", {"wifi_positions": []})

    def get_info(self):
        """Gather information from the system"""
        cur_time = time.time()
        if cur_time - self.last_time > self.update_interval:
            self.last_time = cur_time
            wifi_data = scan_wifi_nmcli()
            use_localconfig = False
            for config in self.lnxlink.config["settings"]["beacondb"]["wifi_positions"]:
                if any(data["ssid"] == config.get("ssid") for data in wifi_data):
                    use_localconfig = True
                    self.position = {
                        "latitude": config.get("latitude"),
                        "longitude": config.get("longitude"),
                        "gps_accuracy": config.get("accuracy", 200),
                    }
            location_result = get_location_from_beacondb(wifi_data, consider_ip=True)
            if location_result and not use_localconfig:
                self.position = {
                    "latitude": location_result.get("location", {}).get("lat"),
                    "longitude": location_result.get("location", {}).get("lng"),
                    "gps_accuracy": location_result.get("accuracy"),
                }
        return self.position

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "BeaconDB": {
                "type": "device_tracker",
                "entity_category": "diagnostic",
            }
        }



"""Information and control of logging level"""
import logging

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Logging Level"
        self.lnxlink = lnxlink

    def exposed_controls(self):
        """Exposes to home assistant"""
        log_levels = [
            "NOTSET",
            "DEBUG",
            "INFO",
            "WARNING",
            "ERROR",
        ]
        return {
            "Logging Level": {
                "type": "select",
                "icon": "mdi:math-log",
                "options": log_levels,
                "entity_category": "diagnostic",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        current_level = logger.getEffectiveLevel()
        level_name = logging.getLevelName(current_level)
        return level_name

    def start_control(self, topic, data):
        """Control system"""
        if data in [0, "notset", "Notset", "NOTSET"]:
            logger.setLevel(logging.NOTSET)
        elif data in [1, 10, "debug", "Debug", "DEBUG"]:
            logger.setLevel(logging.DEBUG)
        elif data in [2, 20, "info", "Info", "INFO"]:
            logger.setLevel(logging.INFO)
        elif data in [3, 30, "warning", "Warning", "WARNING"]:
            logger.setLevel(logging.WARNING)
        elif data in [4, 40, "error", "Error", "ERROR"]:
            logger.setLevel(logging.ERROR)
        elif data in [5, 50, "critical", "Critical", "CRITICAL"]:
            logger.setLevel(logging.CRITICAL)



"""MQTT methods"""

import ssl
import time
import logging
import traceback
import paho.mqtt.client as mqtt

logger = logging.getLogger("lnxlink")


class MQTT:
    """Start LNXlink service that loads all modules and connects to MQTT"""

    def __init__(self, config):
        self.config = config
        self.publish_rc_code = 0

        # Setup mqtt client object
        if hasattr(mqtt, "CallbackAPIVersion"):
            self.client = mqtt.Client(
                client_id=f"LNXlink-{self.config['mqtt']['clientId']}",
                callback_api_version=mqtt.CallbackAPIVersion.VERSION2,
            )
        else:
            self.client = mqtt.Client(
                client_id=f"LNXlink-{self.config['mqtt']['clientId']}"
            )

    def publish(self, topic, payload, retain=True):
        """Publishes messages to the MQTT broker"""
        msg_info = self.client.publish(
            topic,
            payload=payload,
            qos=self.config["mqtt"]["lwt"]["qos"],
            retain=retain,
        )
        logger.debug("Message RC Code: %s, MQTT Number: %s", msg_info.rc, msg_info.mid)
        self.publish_rc_code = msg_info.rc
        # if self.publish_rc_code != 0:
        #     logger.error("Publish RC Code Error, trying to reconnect...")
        #     self.publish_rc_code = 0
        #     self.client.disconnect()
        #     time.sleep(2)
        #     self.client.reconnect()
        #     time.sleep(3)
        return msg_info

    def reconnect(self):
        """Try to reconnect to broker"""
        logger.info("Reconnecting to MQTT")
        self.publish_rc_code = 0
        self.client.disconnect()
        time.sleep(2)
        self.client.reconnect()
        time.sleep(3)

    def disconnect(self):
        """Used when exiting"""
        self.client.disconnect()
        logger.info("Disconnected from MQTT.")
        self.send_lwt("OFF")

    def send_lwt(self, status):
        """Sends the status of lwt, ON or OFF"""
        if status == "OFF" and not self.config["mqtt"]["lwt"]["enabled"]:
            return
        self.client.publish(
            f"{self.config['pref_topic']}/lwt",
            payload=status,
            qos=self.config["mqtt"]["lwt"]["qos"],
            retain=True,
        )

    def setup_mqtt(self, on_connect, on_message):
        """Creates the mqtt object"""
        self.client.on_connect = on_connect
        self.client.on_message = on_message
        self.client.on_disconnect = self.on_disconnect
        self.client.on_publish = self.on_publish

        keyfile = self.config["mqtt"]["auth"]["keyfile"]
        keyfile = None if keyfile == "" else keyfile
        certfile = self.config["mqtt"]["auth"]["certfile"]
        certfile = None if certfile == "" else certfile
        ca_certs = self.config["mqtt"]["auth"]["ca_certs"]
        ca_certs = None if ca_certs == "" else ca_certs
        use_cert = all(option is not None for option in [keyfile, certfile, ca_certs])
        use_tls = self.config["mqtt"]["auth"]["tls"]
        username = self.config["mqtt"]["auth"]["user"]
        password = self.config["mqtt"]["auth"]["pass"]
        use_userpass = all(option != "" for option in [username, password])

        if use_userpass:
            self.client.username_pw_set(username, password)
        if use_tls:
            cert_reqs = ssl.CERT_NONE
            if use_cert:
                cert_reqs = ssl.CERT_REQUIRED
            logger.info("Using MQTT ca_certs: %s", ca_certs)
            logger.info("Using MQTT certfile: %s", certfile)
            logger.info("Using MQTT keyfile: %s", keyfile)
            self.client.tls_set(
                ca_certs=ca_certs,
                certfile=certfile,
                keyfile=keyfile,
                cert_reqs=cert_reqs,
            )
            if ca_certs is None:
                self.client.tls_insecure_set(True)
        if self.config["mqtt"]["lwt"]["enabled"]:
            self.client.will_set(
                f"{self.config['pref_topic']}/lwt",
                payload="OFF",
                qos=self.config["mqtt"]["lwt"]["qos"],
                retain=True,
            )
        try:
            self.client.connect(
                host=self.config["mqtt"]["server"],
                port=self.config["mqtt"]["port"],
                keepalive=60,
            )
        except ssl.SSLCertVerificationError:
            logger.info("TLS not verified, using insecure connection instead")
            self.client.tls_insecure_set(True)
            self.client.connect(
                host=self.config["mqtt"]["server"],
                port=self.config["mqtt"]["port"],
                keepalive=60,
            )
        except Exception as err:
            logger.error(
                "Error establishing connection to MQTT broker: %s, %s",
                err,
                traceback.format_exc(),
            )
            return False
        self.client.loop_start()
        return True

    # pylint: disable=too-many-arguments
    def on_publish(self, client, userdata, mid, reason_code, properties):
        """Trying to reconnect if the reason code is not Success"""
        if reason_code != "Success":
            logger.error("Publish Error, trying to reconnect...")
            self.reconnect()

    def on_disconnect(self, *args):
        """Disconnected from MQTT broker"""
        logger.warning("Lost connection to MQTT Broker...")

    def get_rcode_name(self, rcode):
        """Returns the rcode message"""
        return mqtt.connack_string(rcode)



"""Expose sensors as a Restful API"""
import json
import logging
import threading
import traceback

from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "RESTful"
        self.lnxlink = lnxlink
        self.lnxlink.add_settings("restful", {"port": 8112})
        self._requirements()

    def _requirements(self):
        flask_view = import_install_package("flask", ">=3.0.3", "flask.views")
        flask = import_install_package("flask", ">=3.0.3", "flask")

        class ModuleInfo(flask_view.views.MethodView):
            """Get information from Addon modules"""

            def __init__(self, lnxlink):
                """Init the application"""
                self.lnxlink = lnxlink

            def get(self, module=None):
                """Fetch data from modules"""
                info = self.lnxlink.saved_publish
                if module is None:
                    return json.dumps(list(info.keys()))
                return str(info.get(module))

        class ModuleControl(flask_view.views.MethodView):
            """Control Addon modules"""

            def __init__(self, lnxlink):
                """Init the application"""
                self.lnxlink = lnxlink

            def get(self):
                """Information about control modules"""
                modules = []
                for addonmodule, addon in self.lnxlink.addons.items():
                    if hasattr(addon, "start_control"):
                        modules.append(addonmodule)
                return json.dumps(modules)

            def post(self, module=None):
                """Control an Addon module"""
                if module is None:
                    modules = []
                    for addonmodule, addon in self.lnxlink.addons.items():
                        if hasattr(addon, "start_control"):
                            modules.append(addonmodule)
                    return json.dumps(modules)

                topic = flask.request.form.get("topic", "")
                topic = f"{module}/{topic}"
                topic = topic.split("/")
                message = flask.request.form.get("message")
                addon = self.lnxlink.addons.get(module)
                if addon is not None:
                    if hasattr(addon, "start_control"):
                        try:
                            result = addon.start_control(topic, message)
                            return json.dumps(result)
                        except Exception as err:
                            logger.error(
                                "Couldn't run command for module %s: %s, %s",
                                addon,
                                err,
                                traceback.format_exc(),
                            )
                            return f"Error: {err}"
                    return "No control support available"
                return "Module not found"

        app = flask.Flask(__name__)
        app.add_url_rule(
            "/info", view_func=ModuleInfo.as_view("modules_list", self.lnxlink)
        )
        app.add_url_rule(
            "/info/<module>", view_func=ModuleInfo.as_view("module_info", self.lnxlink)
        )
        app.add_url_rule(
            "/control", view_func=ModuleControl.as_view("control_list", self.lnxlink)
        )
        app.add_url_rule(
            "/control/<module>",
            view_func=ModuleControl.as_view("control", self.lnxlink),
        )
        threading.Thread(target=self._serve, args=[app], daemon=True).start()

    def _serve(self, app):
        waitress = import_install_package("waitress", ">=3.0.0", "waitress")
        port = self.lnxlink.config["settings"].get("restful", {}).get("port", 8112)
        waitress.serve(app, host="0.0.0.0", port=port)



"""Gets the battery information of connected devices"""
from xml.etree import ElementTree
from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Battery"
        self.lnxlink = lnxlink
        self.conn = open_dbus_connection(bus="SYSTEM")
        self.lnxlink.add_settings(
            "battery",
            {
                "include_batteries": [],
                "exclude_batteries": [],
            },
        )
        self.devices = self._get_devices()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.devices:
            att_temp = f"{{{{ value_json.get('{device}', {{}}).get('attributes', {{}}) | tojson }}}}"
            discovery_info[f"Battery {device}"] = {
                "type": "sensor",
                "unit": "%",
                "device_class": "battery",
                "value_template": f"{{{{ value_json.get('{device}', {{}}).get('percent') }}}}",
                "attributes_template": att_temp,
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        devices = self._get_devices()
        new_devices = set(devices) - set(self.devices)
        disconnected_devices = set(self.devices) - set(devices)
        for device_name in disconnected_devices:
            devices[device_name] = self.devices[device_name]
            self.devices[device_name]["percent"] = None
        self.devices = devices
        if len(new_devices) > 0:
            self.lnxlink.setup_discovery("battery")

        return devices

    def _get_devices(self):
        devices = {}
        battery_includes = (
            self.lnxlink.config["settings"]
            .get("battery", {})
            .get("include_batteries", [])
        )
        battery_excludes = (
            self.lnxlink.config["settings"]
            .get("battery", {})
            .get("exclude_batteries", [])
        )

        u_power_states = {
            0: "unknown",
            1: "charging",
            2: "discharging",
            3: "empty",
            4: "fully charged",
            5: "pending charge",
            6: "pending discharge",
        }

        for device in self.get_batteries():
            if len(battery_includes) != 0:
                if not any(device["Model"].startswith(x) for x in battery_includes):
                    continue
            if len(battery_excludes) != 0:
                if any(device["Model"].startswith(x) for x in battery_excludes):
                    continue

            native_path = device["NativePath"].split("/")[-1]
            name = (
                " ".join(
                    [
                        device["Vendor"],
                        device["Model"],
                        device["Serial"].replace(":", "").replace("/", ""),
                    ]
                )
                .strip()
                .replace("'", "_")
            )
            if name == "":
                name = native_path
            if name != "":
                devices[name] = {
                    "percent": device["Percentage"],
                    "attributes": {
                        "vendor": device["Vendor"],
                        "model": device["Model"],
                        "serial": device["Serial"],
                        "native_path": native_path,
                        "rechargeable": device["IsRechargeable"],
                        "status": u_power_states[device["State"]],
                        "time_to_empty": device["TimeToEmpty"],
                        "time_to_full": device["TimeToFull"],
                    },
                }
        return devices

    def dbus_paths(self, service, object_path, paths):
        """Recursively get all child object paths via introspection"""
        introspect_iface = "org.freedesktop.DBus.Introspectable"
        addr = DBusAddress(object_path, bus_name=service, interface=introspect_iface)
        msg = new_method_call(addr, "Introspect")
        reply = self.conn.send_and_get_reply(msg)
        xml_string = reply.body[0]

        for child in ElementTree.fromstring(xml_string):
            if child.tag == "node":
                name = child.attrib["name"]
                new_path = object_path.rstrip("/") + "/" + name
                paths.append(new_path)
                self.dbus_paths(service, new_path, paths)
        return paths

    def get_property(self, object_path, interface, prop):
        """Gets the device property"""
        addr = DBusAddress(
            object_path,
            bus_name="org.freedesktop.UPower",
            interface="org.freedesktop.DBus.Properties",
        )
        msg = new_method_call(addr, "Get", "ss", (interface, prop))
        reply = self.conn.send_and_get_reply(msg)
        return reply.body[0][1]

    def get_batteries(self):
        """Gets a list of all devices and their status"""
        device_iface = "org.freedesktop.UPower.Device"
        batteries = []
        paths = self.dbus_paths("org.freedesktop.UPower", "/org/freedesktop/UPower", [])

        for path in paths:
            try:
                # Check if it's a UPower.Device
                props = {
                    "Model": self.get_property(path, device_iface, "Model"),
                    "NativePath": self.get_property(path, device_iface, "NativePath"),
                    "Percentage": self.get_property(path, device_iface, "Percentage"),
                    "Serial": self.get_property(path, device_iface, "Serial"),
                    "IconName": self.get_property(path, device_iface, "IconName"),
                    "IsRechargeable": self.get_property(
                        path, device_iface, "IsRechargeable"
                    ),
                    "Vendor": self.get_property(path, device_iface, "Vendor"),
                    "State": self.get_property(path, device_iface, "State"),
                    "TimeToEmpty": self.get_property(path, device_iface, "TimeToEmpty"),
                    "TimeToFull": self.get_property(path, device_iface, "TimeToFull"),
                }
                if props["Model"] + props["NativePath"] not in ["bb", "b", ""]:
                    if isinstance(props["Percentage"], float):
                        batteries.append(props)
            except Exception:
                # Not a UPower.Device or missing properties
                continue

        return batteries



"""Controls Docker instance"""
import time
import logging
from lnxlink.modules.scripts.docker_update_status import DockerUpdateStatus
from lnxlink.modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    # pylint: disable=too-many-branches
    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Docker"
        self.lnxlink = lnxlink
        self.docker = import_install_package("docker", ">=7.0.0")
        if self.docker is None:
            raise SystemError("Docker package not found")
        try:
            # client = docker.from_env()
            self.client = self.docker.DockerClient(base_url="unix://run/docker.sock")
        except Exception as err:
            raise SystemError(f"Docker instance not found: {err}") from err
        self.lnxlink.add_settings(
            "docker",
            {
                "include": [],
                "exclude": [],
                "check_update": 24,
                "expose_controls": True,
            },
        )
        self.prev_update = 0
        self.images_remoteinfo = []
        self.containers = self._get_containers()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for container in self.containers:
            attr_templ = f"{{{{ value_json.get('{container}', {{}}).get('attrs', {{}}) | tojson }}}}"
            if self.lnxlink.config["settings"]["docker"]["expose_controls"]:
                discovery_info[f"Docker {container}"] = {
                    "type": "switch",
                    "icon": "mdi:docker",
                    "value_template": f"{{{{ value_json.get('{container}', {{}}).get('running') }}}}",
                    "attributes_template": attr_templ,
                }
            if self.lnxlink.config["settings"]["docker"]["check_update"] is not None:
                discovery_info[f"Docker Update {container}"] = {
                    "type": "binary_sensor",
                    "icon": "mdi:docker",
                    "value_template": f"{{{{ value_json.get('{container}', {{}}).get('update') }}}}",
                }
        discovery_info["Docker Prune"] = {
            "type": "button",
            "icon": "mdi:docker",
        }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        containers = self._get_containers()
        if len(containers) != len(self.containers):
            self.lnxlink.setup_discovery("docker")
        self.containers = containers
        return self.containers

    def _get_containers(self):
        include = self.lnxlink.config["settings"].get("docker", {}).get("include", [])
        exclude = self.lnxlink.config["settings"].get("docker", {}).get("exclude", [])
        containers = {}
        images = []
        for container in self.client.containers.list(all=True):
            if len(include) > 0 and container.name not in include:
                continue
            if container.name in exclude:
                continue
            ports = set()
            for _, host in container.ports.items():
                if host is not None:
                    for host_info in host:
                        ports.add(host_info["HostPort"])
            running = "OFF"
            if container.attrs["State"]["Running"]:
                running = "ON"
            name_id = container.name.lower().replace(" ", "_")
            images.append(container.image)
            containers[name_id] = {
                "running": running,
                "attrs": {
                    "name": container.name,
                    "images": ",".join(container.image.tags),
                    "ports": list(ports),
                    "status": container.status,
                    "update": None,
                },
            }

        cur_time = time.time() / 60 / 60
        check_update = self.lnxlink.config["settings"]["docker"]["check_update"]
        if check_update is not None:
            if cur_time - self.prev_update > check_update:
                self.prev_update = cur_time
                docker_update_status = DockerUpdateStatus()
                self.images_remoteinfo = docker_update_status.get_updates_sync(images)

            for remoteimage_info in self.images_remoteinfo:
                for container in containers.values():
                    if remoteimage_info["tag"] in container["attrs"]["images"]:
                        if remoteimage_info["status"] == "update_available":
                            container["attrs"]["update"] = True
                            container["update"] = "ON"
                        elif remoteimage_info["status"] == "up_to_date":
                            container["attrs"]["update"] = False
                            container["update"] = "OFF"

        return containers

    def start_control(self, topic, data):
        """Control system"""
        name_id = topic[1].replace("docker_", "")
        if name_id in self.containers:
            name = self.containers[name_id]["attrs"]["name"]
            if data == "ON":
                logger.info("Starting container %s", name)
                self.client.containers.get(name).start()
            elif data == "OFF":
                logger.info("Stopping container %s", name)
                self.client.containers.get(name).stop()
        elif name_id == "prune":
            # docker system prune -af
            logger.info("Running prune all")
            self.client.containers.prune()
            self.client.images.prune()
            self.client.networks.prune()
            self.client.volumes.prune()



"""Checks for updates on selected docker images"""

import os
from typing import Optional, Dict, List, Tuple, Any
from urllib.parse import quote_plus
import asyncio
import aiohttp

# --- Configuration ---
GITLAB_TOKEN = os.getenv("GITLAB_TOKEN")
QUAY_TOKEN = os.getenv("QUAY_TOKEN")


# --- Registry Clients ---


class RegistryClient:
    """Base class for a registry client."""

    def __init__(self, repo: str, session: aiohttp.ClientSession):
        self.repo = repo
        self.session = session
        self.base_url = ""

    async def get_remote_digest(self, tag: str) -> Optional[str]:
        """Fetches the manifest for a tag and returns its digest."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    @staticmethod
    def parse_repo_name(image_name: str) -> Tuple[str, str, str]:
        """Parses an image name into (registry, repository, tag)."""
        if ":" not in image_name:
            image_name += ":latest"
        registry_part, tag = image_name.rsplit(":", 1)
        parts = registry_part.split("/")
        registry = "docker.io"
        if "." in parts[0] and len(parts) > 1:
            registry = parts[0]
            repository = "/".join(parts[1:])
        else:
            repository = registry_part
            if "/" not in repository:
                repository = f"library/{repository}"
        return registry, repository, tag


class DockerHubClient(RegistryClient):
    """Client for Docker Hub."""

    def __init__(self, repo: str, session: aiohttp.ClientSession):
        super().__init__(repo, session)
        self.auth_url = "https://auth.docker.io/token?service=registry.docker.io"
        self.base_url = "https://registry-1.docker.io/v2"

    async def _get_auth_token(self) -> str:
        url = f"{self.auth_url}&scope=repository:{self.repo}:pull"
        async with self.session.get(url) as response:
            response.raise_for_status()
            data = await response.json()
            return data["token"]

    async def get_remote_digest(self, tag: str) -> Optional[str]:
        try:
            token = await self._get_auth_token()
            headers = {
                "Authorization": f"Bearer {token}",
                "Accept": "application/vnd.docker.distribution.manifest.v2+json, application/vnd.oci.image.index.v1+json",  # noqa: E501
            }
            url = f"{self.base_url}/{self.repo}/manifests/{tag}"
            async with self.session.head(url, headers=headers) as response:
                response.raise_for_status()
                return response.headers.get("Docker-Content-Digest")
        except aiohttp.ClientError:
            return None


class GhcrClient(RegistryClient):
    """Client for GitHub Container Registry (ghcr.io)."""

    def __init__(self, repo: str, session: aiohttp.ClientSession):
        super().__init__(repo, session)
        self.base_url = "https://ghcr.io/v2"
        self.token_url = (
            f"https://ghcr.io/token?service=ghcr.io&scope=repository:{self.repo}:pull"
        )

    async def _get_anonymous_token(self) -> Optional[str]:
        """Fetches a temporary public token for GHCR."""
        try:
            async with self.session.get(self.token_url) as response:
                response.raise_for_status()
                data = await response.json()
                return data["token"]
        except aiohttp.ClientError:
            return None

    async def get_remote_digest(self, tag: str) -> Optional[str]:
        headers = {
            "Accept": "application/vnd.docker.distribution.manifest.v2+json, application/vnd.oci.image.index.v1+json"  # noqa: E501
        }
        token = await self._get_anonymous_token()
        if not token:
            # We return None and let the caller decide how to report the error
            return None
        headers["Authorization"] = f"Bearer {token}"
        try:
            url = f"{self.base_url}/{self.repo}/manifests/{tag}"
            async with self.session.head(url, headers=headers) as response:
                response.raise_for_status()
                return response.headers.get("Docker-Content-Digest")
        except aiohttp.ClientError:
            return None


class LscrClient(GhcrClient):
    """Client for LinuxServer Container Registry (lscr.io)."""

    def __init__(self, repo: str, session: aiohttp.ClientSession):
        super().__init__(repo, session)
        self.base_url = "https://lscr.io/v2"


class QuayClient(RegistryClient):
    """Client for Quay.io."""

    def __init__(self, repo: str, session: aiohttp.ClientSession):
        super().__init__(repo, session)
        self.base_url = "https://quay.io/api/v1"

    async def get_remote_digest(self, tag: str) -> Optional[str]:
        headers = {"Accept": "application/json"}
        if QUAY_TOKEN:
            headers["Authorization"] = f"Bearer {QUAY_TOKEN}"
        try:
            url = f"{self.base_url}/repository/{self.repo}/tag/{tag}"
            async with self.session.get(url, headers=headers) as response:
                response.raise_for_status()
                data = await response.json()
                return data.get("manifest_digest")
        except aiohttp.ClientError:
            return None


class GitLabClient(RegistryClient):
    """Client for GitLab Container Registry."""

    def __init__(self, repo: str, session: aiohttp.ClientSession):
        super().__init__(repo, session)
        self.base_url = "https://gitlab.com/api/v4"
        self._project_id = None

    async def _get_project_id(self) -> Optional[str]:
        if self._project_id:
            return self._project_id
        if not GITLAB_TOKEN:
            return None
        project_path = (
            "/".join(self.repo.split("/")[:-1]) if "/" in self.repo else self.repo
        )
        try:
            headers = {"PRIVATE-TOKEN": GITLAB_TOKEN}
            encoded_project_path = quote_plus(project_path)
            url = f"{self.base_url}/projects/{encoded_project_path}"
            async with self.session.get(url, headers=headers) as response:
                response.raise_for_status()
                data = await response.json()
                self._project_id = data["id"]
                return self._project_id
        except aiohttp.ClientError:
            return None

    async def get_remote_digest(self, tag: str) -> Optional[str]:
        project_id = await self._get_project_id()
        if not project_id:
            return None
        try:
            headers = {"PRIVATE-TOKEN": GITLAB_TOKEN}
            repo_url = f"{self.base_url}/projects/{project_id}/registry/repositories"
            async with self.session.get(repo_url, headers=headers) as repo_response:
                repo_response.raise_for_status()
                repos_data = await repo_response.json()
                repo_id = next(
                    (repo["id"] for repo in repos_data if repo["path"] == self.repo),
                    None,
                )
            if not repo_id:
                return None
            tag_url = f"{self.base_url}/projects/{project_id}/registry/repositories/{repo_id}/tags/{tag}"
            async with self.session.get(tag_url, headers=headers) as tag_response:
                tag_response.raise_for_status()
                tag_data = await tag_response.json()
                return tag_data["digest"]
        except aiohttp.ClientError:
            return None


class DockerUpdateStatus:
    """docstring for DockerUpdateStatus"""

    def get_registry_client(
        self, registry: str, repo: str, session: aiohttp.ClientSession
    ) -> Optional[RegistryClient]:
        """Factory function to return the appropriate client for a given registry."""
        if registry == "docker.io":
            return DockerHubClient(repo, session)
        if registry == "ghcr.io":
            return GhcrClient(repo, session)
        if registry == "lscr.io":
            return LscrClient(repo, session)
        if registry == "quay.io":
            return QuayClient(repo, session)
        if registry.endswith("gitlab.com"):
            return GitLabClient(repo, session)
        return None

    # pylint: disable=no-else-return
    async def check_image_tag(
        self,
        full_tag: str,
        image_id: str,
        repo_digests: List[str],
        session: aiohttp.ClientSession,
    ) -> Dict[str, Any]:
        """
        Checks a single image tag and returns a result dictionary.
        """
        try:
            registry, repo, tag = RegistryClient.parse_repo_name(full_tag)
            local_digest = ""
            if repo_digests:
                for d in repo_digests:
                    if d.startswith(repo) or d.startswith(f"{registry}/{repo}"):
                        local_digest = d.split("@")[1]
                        break
            if not local_digest:
                local_digest = image_id

            client = self.get_registry_client(registry, repo, session)
            if not client:
                return {
                    "tag": full_tag,
                    "status": "error",
                    "message": f"Unsupported registry: {registry}",
                }

            remote_digest = await client.get_remote_digest(tag)

            if not remote_digest:
                return {
                    "tag": full_tag,
                    "status": "error",
                    "message": "Could not fetch remote digest.",
                }
            elif local_digest.strip() == remote_digest.strip():
                return {
                    "tag": full_tag,
                    "status": "up_to_date",
                }
            else:
                return {
                    "tag": full_tag,
                    "status": "update_available",
                    "local": local_digest,
                    "remote": remote_digest,
                }
        except Exception as e:
            return {
                "tag": full_tag,
                "status": "error",
                "message": str(e),
            }

    async def get_updates(self, images):
        """Checks for updates async"""
        async with aiohttp.ClientSession() as session:
            tasks = []
            for image in images:
                if not image.tags:
                    continue
                for tag in image.tags:
                    task = self.check_image_tag(
                        tag, image.id, image.attrs.get("RepoDigests", []), session
                    )
                    tasks.append(task)

            all_results = await asyncio.gather(*tasks)

        return all_results

    def get_updates_sync(self, images):
        """Run the get_updates synchronously"""
        return asyncio.run(self.get_updates(images))



"""Restarts the system"""
import logging
from shutil import which
from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Restart"
        self.lnxlink = lnxlink

    def start_control(self, topic, data):
        """Control system"""
        self.lnxlink.temp_connection_callback(True)
        returncode = None
        if which("systemctl") is not None and returncode != 0:
            _, _, returncode = syscommand("systemctl reboot")
        if which("shutdown") is not None and returncode != 0:
            _, _, returncode = syscommand("shutdown -r now")
        if returncode != 0:
            try:
                conn = open_dbus_connection(bus="SYSTEM")
                conn.send(
                    new_method_call(
                        DBusAddress(
                            object_path="/org/freedesktop/login1",
                            bus_name="org.freedesktop.login1",
                            interface="org.freedesktop.login1.Manager",
                        ),
                        method="Reboot",
                        signature="b",
                        body=(True,),
                    )
                )
            except Exception:
                returncode = -1
        if returncode != 0:
            self.lnxlink.temp_connection_callback(False)

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Restart": {
                "type": "button",
                "icon": "mdi:restart",
            }
        }



"""Controls Paired Bluetooth Devices"""
import re
import logging
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Bluetooth"
        self.lnxlink = lnxlink
        _, _, returncode = syscommand("bluetoothctl show")
        if returncode != 0:
            raise SystemError("Can't run command bluetoothctl")
        self.bluetoothdata = self._get_bluetoothdata()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {
            "Bluetooth Power": {
                "type": "switch",
                "icon": "mdi:bluetooth-connect",
                "value_template": "{{ value_json.get('power') }}",
            }
        }
        for mac, blinfo in self.bluetoothdata["devices"].items():
            attr_templ = f"{{{{ value_json.devices.get('{mac}', {{}}).get('attributes') | tojson }}}}"
            discovery_info[
                f"Bluetooth Device {blinfo['name'].replace('+', '')} {mac.replace(':', '')}"
            ] = {
                "type": "switch",
                "icon": "mdi:bluetooth",
                "value_template": f"{{{{ value_json.devices.get('{mac}', {{}}).get('power') }}}}",
                "attributes_template": attr_templ,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        bluetoothdata = self._get_bluetoothdata()
        loaded = set(bluetoothdata["devices"]) - set(self.bluetoothdata["devices"])
        unloaded = set(self.bluetoothdata["devices"]) - set(bluetoothdata["devices"])
        for device in unloaded:
            bluetoothdevices = self.bluetoothdata.get(device)
            if bluetoothdevices is not None:
                for addr in bluetoothdevices:
                    if addr is not None:
                        self.bluetoothdata[device][addr] = None
            if device in self.bluetoothdata:
                bluetoothdata[device] = self.bluetoothdata[device]
        self.bluetoothdata = bluetoothdata
        if len(loaded) > 0:
            self.lnxlink.setup_discovery("bluetooth")
        return self.bluetoothdata

    def _bytetogb(self, byte):
        return round(byte / 1024 / 1024 / 1024, 0)

    def _get_bluetoothdata(self):
        """Get a list of all bluetooth devices"""
        data = {
            "power": "OFF",
            "devices": {},
            "attributes": {
                "battery": None,
            },
        }
        stdout, _, _ = syscommand("bluetoothctl show | grep Powered")
        if "yes" in stdout:
            data["power"] = "ON"

        stdout, _, returncode = syscommand(
            "bluetoothctl devices Paired",
            ignore_errors=True,
        )
        if returncode != 0:
            stdout, _, _ = syscommand(
                "bluetoothctl paired-devices",
                ignore_errors=True,
            )
        if stdout == "":
            return data
        for device in stdout.split("\n"):
            try:
                device_type, mac, name = device.split(" ", maxsplit=2)
                if device_type != "Device":
                    continue
            except Exception:
                continue
            stdoutdevice, _, _ = syscommand(f"bluetoothctl info {mac}")
            power = "OFF"
            if re.search(r"Connected:\s*yes", stdoutdevice):
                power = "ON"
            battery = None
            match = re.search(r"Battery Percentage:.*\((\d+)\)", stdoutdevice)
            if match:
                battery = match.group(1)
            data["devices"][mac] = {
                "name": name,
                "power": power,
                "attributes": {
                    "battery": battery,
                },
            }

        return data

    def start_control(self, topic, data):
        """Control system"""
        if topic[1] == "bluetooth_power":
            syscommand(f"bluetoothctl power {data.lower()}")
        else:
            simple_mac = topic[1].split("_")[-1]
            mac = ":".join(re.findall("..?", simple_mac))
            if data == "ON":
                syscommand(f"bluetoothctl connect {mac}")
            elif data == "OFF":
                syscommand(f"bluetoothctl disconnect {mac}")



"""Gets Display Environment"""
from lnxlink.modules.scripts.helpers import get_display_variable


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Display Environment"

    def get_info(self):
        """Gather information from the system"""
        return get_display_variable()

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Display Environment": {
                "type": "sensor",
                "icon": "mdi:panorama-variant-outline",
                "entity_category": "diagnostic",
                "enabled": False,
            }
        }



"""Run a terminal command"""
import logging
import time
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "bash"
        self.lnxlink = lnxlink
        self.discovery_info = {}
        self.lnxlink.add_settings(
            "bash",
            {
                "allow_any_command": False,
                "expose": [],
            },
        )

    def get_info(self):
        """Gather information from the system"""
        for expose_name, discovery in self.discovery_info.items():
            if discovery.get("type") == "sensor":
                cur_time = time.time()
                if cur_time - discovery["last_time"] > discovery["update_interval"]:
                    discovery["last_time"] = cur_time
                    stdout, _, returncode = syscommand(
                        discovery["local_command"],
                        timeout=discovery.get("sensor_timeout", 3),
                    )
                    if returncode == 0:
                        self.lnxlink.run_module(f"{self.name}/{expose_name}", stdout)
            elif discovery.get("type") == "binary_sensor":
                cur_time = time.time()
                if cur_time - discovery["last_time"] > discovery["update_interval"]:
                    discovery["last_time"] = cur_time
                    stdout, _, _ = syscommand(
                        discovery["local_command"],
                        timeout=discovery.get("sensor_timeout", 3),
                    )
                    status = stdout.lower() not in ["false", "no", "0", "off", ""]
                    senddata = {
                        "status": "ON" if status else "OFF",
                        "attributes": {"raw": stdout.split("\n")},
                    }
                    self.lnxlink.run_module(f"{self.name}/{expose_name}", senddata)
            elif discovery.get("type") == "switch":
                stdout, _, _ = syscommand(
                    discovery["local_command"],
                    ignore_errors=True,
                    timeout=discovery.get("sensor_timeout", 3),
                )
                status = stdout.lower() not in ["false", "no", "0", "off", ""]
                self.lnxlink.run_module(f"{self.name}/{expose_name}", status)

    def exposed_controls(self):
        """Exposes to home assistant"""
        self.discovery_info = {}
        if self.lnxlink.config["settings"]["bash"]["allow_any_command"]:
            self.discovery_info["Bash Command"] = {
                "type": "text",
                "icon": "mdi:bash",
            }

        exposed = self.lnxlink.config["settings"]["bash"]["expose"]
        exposed = [] if exposed is None else exposed
        for expose in exposed:
            expose_type = expose.get("type", "button")
            expose_name = f"Bash {expose['name']}"
            icon = expose.get("icon", "mdi:script-text")
            if expose_type == "button":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "payload_press": expose["command"],
                }
            elif expose_type == "sensor":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "unit": expose.get("unit"),
                    "local_command": expose.get("command"),
                    "subtopic": True,
                    "update_interval": expose.get("update_interval", 0),
                    "last_time": 0,
                    "sensor_timeout": expose.get("sensor_timeout", 3),
                }
            elif expose_type == "binary_sensor":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "value_template": "{{ value_json.status }}",
                    "attributes_template": "{{ value_json.attributes | tojson }}",
                    "local_command": expose.get("command"),
                    "subtopic": True,
                    "update_interval": expose.get("update_interval", 0),
                    "last_time": 0,
                    "sensor_timeout": expose.get("sensor_timeout", 3),
                }
            elif expose_type == "switch":
                self.discovery_info[expose_name] = {
                    "type": expose_type,
                    "icon": icon,
                    "local_command": expose.get("command"),
                    "command_on": expose.get("command_on"),
                    "command_off": expose.get("command_off"),
                    "subtopic": True,
                    "sensor_timeout": expose.get("sensor_timeout", 3),
                }
            if expose.get("entity_category") in ["diagnostic", "config"]:
                self.discovery_info[expose_name]["entity_category"] = expose[
                    "entity_category"
                ]

        return self.discovery_info

    def start_control(self, topic, data):
        """Control system"""
        allow_any_command = self.lnxlink.config["settings"]["bash"]["allow_any_command"]
        if allow_any_command:
            stdout, _, _ = syscommand(data, timeout=120)
            return stdout
        exposed = self.lnxlink.config["settings"]["bash"]["expose"]
        exposed = [] if exposed is None else exposed
        for expose in exposed:
            command_list = [
                expose.get("command", "").strip(),
                expose.get("command_on", "").strip(),
                expose.get("command_off", "").strip(),
            ]
            if data.strip() in command_list:
                stdout, _, _ = syscommand(
                    data,
                    timeout=expose.get("command_timeout", 120),
                )
                return stdout
        logger.error(
            "Check bash configuration option allow_any_command to run this command: %s",
            data,
        )
        return None



"""Helper functions to get information from files"""
import os
import time
import logging
import importlib.metadata
from logging.handlers import RotatingFileHandler
from collections import OrderedDict

import yaml
from lnxlink.modules.scripts.helpers import syscommand


logger = logging.getLogger("lnxlink")


class UniqueQueue:
    """
    A queue that maintains unique named items with a maximum size limit.
    If an item with the same name is added, it replaces the old one.
    When full, the oldest item is discarded to make room.
    """

    def __init__(self, max_size=200):
        """Initializes the UniqueQueue"""
        self.queue = OrderedDict()
        self.max_size = max_size

    def __repr__(self):
        """Returns a string representation of the queue."""
        return f"<{self.__class__.__name__} queue: {repr(self.queue)}>"

    def __iter__(self):
        """Returns an iterator that yields and removes items from the queue in FIFO order"""
        while self.queue:
            yield self.queue.popitem(last=False)

    def add_item(self, name, value, retain=True):
        """Adds an item to the queue. If the item already exists, it replaces it"""
        # If item exists, remove it so we can re-add at the end
        if name in self.queue:
            del self.queue[name]
        # If queue is full, remove the oldest item
        elif len(self.queue) >= self.max_size:
            self.queue.popitem(last=False)
        self.queue[name] = (value, retain)

    def get_item(self):
        """Retrieves and removes the next item from the queue (FIFO)"""
        if self.queue:
            return self.queue.popitem(last=False)
        return None, None

    def clear(self):
        """Clears all items from the queue"""
        self.queue.clear()


def setup_logger(config_path, log_level):
    """Save logs on the same directory as the config file"""
    config_dir = os.path.dirname(os.path.realpath(config_path))
    start_sec = str(int(time.time()))[-4:]
    log_formatter = logging.Formatter(
        "%(asctime)s ["
        + start_sec
        + ":%(threadName)s.%(module)s.%(funcName)s.%(lineno)d] [%(levelname)s]  %(message)s"
    )

    file_handler = RotatingFileHandler(
        f"{config_dir}/lnxlink.log",
        maxBytes=5 * 1024 * 1024,
        backupCount=1,
    )
    logging.basicConfig(level=log_level)
    file_handler.setFormatter(log_formatter)
    logger.addHandler(file_handler)


def read_config(config_path):
    """Reads the config file and prepares module names for import"""
    with open(config_path, "r", encoding="utf8") as file:
        conf = yaml.load(file, Loader=yaml.FullLoader)

    pref_topic = f"{conf['mqtt']['prefix']}/{conf['mqtt']['clientId']}"
    conf["pref_topic"] = pref_topic.lower()
    conf["config_path"] = config_path

    if conf["modules"] is not None:
        conf["modules"] = [x.lower().replace("-", "_") for x in conf["modules"]]

    if os.environ.get("LNXLINK_MQTT_PREFIX") not in [None, ""]:
        conf["mqtt"]["prefix"] = os.environ.get("LNXLINK_MQTT_PREFIX")
    if os.environ.get("LNXLINK_MQTT_CLIENTID") not in [None, ""]:
        conf["mqtt"]["clientId"] = os.environ.get("LNXLINK_MQTT_CLIENTID")
    if os.environ.get("LNXLINK_MQTT_SERVER") not in [None, ""]:
        conf["mqtt"]["server"] = os.environ.get("LNXLINK_MQTT_SERVER")
    if os.environ.get("LNXLINK_MQTT_PORT") not in [None, ""]:
        conf["mqtt"]["port"] = os.environ.get("LNXLINK_MQTT_PORT")
    if os.environ.get("LNXLINK_MQTT_USER") not in [None, ""]:
        conf["mqtt"]["auth"]["user"] = os.environ.get("LNXLINK_MQTT_USER")
    if os.environ.get("LNXLINK_MQTT_PASS") not in [None, ""]:
        conf["mqtt"]["auth"]["pass"] = os.environ.get("LNXLINK_MQTT_PASS")

    return conf


def get_version():
    """Get the current version and the path of the app"""
    version = importlib.metadata.version(__package__ or __name__)
    path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    if os.path.exists(os.path.join(path, "lnxlink/edit.txt")):
        version += "+edit"
        git_hash, _, return_code = syscommand(
            f"git -C {path} rev-parse --short HEAD",
            ignore_errors=True,
        )
        if return_code == 0:
            version += f"-{git_hash}"
    return version, path



"""Assign hotkeys to run commands"""
import logging
from datetime import datetime
from lnxlink.modules.scripts.helpers import import_install_package, get_display_variable


logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Keyboard Hotkeys"
        self.lnxlink = lnxlink
        self.started = False
        self.lib = {}
        self._requirements()
        self.lnxlink.add_settings("hotkeys", [])
        if len(self.lnxlink.config["settings"]["hotkeys"]) == 0:
            logger.warning("No hotkeys configured for Keyboard Hotkeys module")
            logger.warning("  Please add hotkeys in the settings like so:")
            logger.warning("  - key: <ctrl>+<alt>+a")

    def exposed_controls(self):
        """Exposes to home assistant"""
        use_sensor = any(
            "type" not in v for v in self.lnxlink.config["settings"]["hotkeys"]
        )
        discovery_info = {}
        if use_sensor:
            discovery_info = {
                "Keyboard Hotkey": {
                    "type": "sensor",
                    "device_class": "timestamp",
                    "value_template": "{{ value_json.timevalue }}",
                    "attributes_template": "{{ value_json.attributes | tojson }}",
                },
            }
        return discovery_info

    def _requirements(self):
        self.lib = {
            "xlib_hotkeys": import_install_package(
                "xlib-hotkeys", ">=2024.3.0", "xlib_hotkeys"
            ),
        }

    def get_info(self):
        """Gather information from the system"""
        display_variable = get_display_variable()
        if not self.started and display_variable is not None:
            hm = self.lib["xlib_hotkeys"].HotKeysManager(display_variable)
            actions = self.lnxlink.config["settings"]["hotkeys"]
            for action in actions:
                if isinstance(action, str):
                    action = {"key": action}
                action_key = action.get("key")
                hm.hotkeys[action_key] = lambda action=action: self._activate(action)
            hm.start()
            self.started = True

    def _activate(self, act):
        action = act.copy()
        key = action.pop("key")
        data_send = {
            "timevalue": datetime.now().astimezone().isoformat(),
            "attributes": {
                "key": key,
            },
        }
        self.lnxlink.run_module(self.name, data_send)



"""Uses xdotool to press keyboard keys"""
import os
import re
import logging
from shutil import which
from lnxlink.modules.scripts.helpers import syscommand, get_display_variable

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Send Keys"
        self.used_tool = None
        if which("ydotool") is not None:
            self.used_tool = "ydotool"
        elif which("xdotool") is not None:
            self.used_tool = "xdotool"
        else:
            raise SystemError("System command 'xdotool' or 'ydotool' not found")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Send Keys": {
                "type": "text",
                "icon": "mdi:keyboard-outline",
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        display_variable = get_display_variable()
        if display_variable is not None and os.environ["DISPLAY"] is None:
            os.environ["DISPLAY"] = display_variable
            logger.info("Initializing empty DISPLAY environment variable")
        if self.used_tool == "ydotool":
            event_codes_path = "/usr/include/linux/input-event-codes.h"
            key_map = self._extract_key_definitions(event_codes_path)
            data = self._create_key_representation(data, key_map)
            logger.debug("Sending keys via ydotool: %s", data)
        syscommand(f"{self.used_tool} key {data}")

    def _create_key_representation(self, key_string, key_map):
        """
        Converts a string like "<ctrl>+<alt>+a+b" into a list of press/release tuples.
        """
        special_keys = ["ctrl", "alt", "shift", "cmd", "super"]
        cleaned_string = key_string.lower().replace(" ", "")
        keys = cleaned_string.split("+")

        sequence = []
        modifiers_pressed_in_order = []

        for key in keys:
            if key in special_keys or (key.startswith("<") and key.endswith(">")):
                key = key.replace("<", "").replace(">", "")
                sequence.append((key, 1))
                modifiers_pressed_in_order.append(key)
            else:
                sequence.append((key, 1))
                sequence.append((key, 0))
        for key in reversed(modifiers_pressed_in_order):
            sequence.append((key, 0))

        logger.debug("press sequence for ydotool: %s", sequence)
        formatted_parts = []
        for key_name, action in sequence:
            if key_name in key_map:
                code = key_map.get(key_name)
                formatted_string = ":".join(map(str, [code, action]))
                formatted_parts.append(formatted_string)
            else:
                logger.warning("Key '%s' not found in key map.", key_name)
        formatted_output = " ".join(formatted_parts)

        return formatted_output

    def _extract_key_definitions(self, file_path):
        """
        Reads a file, finds lines starting with '#define KEY_', and extracts
        the key name and its integer/hexadecimal value.
        """
        key_definitions = {}

        # Pattern remains the same: capture KEY_name and the value
        pattern = r"^\s*#define\s+KEY_([^\s]+)\s+([0-9xXA-Fa-f]+).*$"

        # Check if the file exists before attempting to open it
        if not os.path.exists(file_path):
            logger.error("The file was not found at '%s'.", file_path)
            return None

        try:
            with open(file_path, "r", encoding="utf-8") as f:
                for line_number, line in enumerate(f, 1):
                    match = re.search(pattern, line)

                    if match:
                        key_name = match.group(1).lower()
                        value_str = match.group(2)

                        try:
                            if value_str.lower().startswith("0x"):
                                value = int(value_str, 16)
                            else:
                                value = int(value_str)

                            key_definitions[key_name] = value
                            if key_name.startswith("left") and not key_name.endswith(
                                "left"
                            ):
                                key_name_dup = key_name.replace("left", "")
                                key_definitions[key_name_dup] = value
                        except ValueError:
                            logger.warning(
                                "Failed to convert value '%s' on line %s in file '%s'",
                                value_str,
                                line_number,
                                file_path,
                            )

        except IOError as e:
            logger.error(
                "An I/O error occurred while reading the file '%s': %s", file_path, e
            )
            return None

        return key_definitions



"""Controls systemd services"""
import logging
from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection
from lnxlink.modules.scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "SystemD"
        self.lnxlink = lnxlink
        self.lnxlink.add_settings("systemd", [])
        self.services = self.lnxlink.config["settings"].get("systemd", [])
        self.services = [] if self.services is None else self.services
        if not self.services:
            logger.info("No systemd settings found on configuration.")
        self.bus = open_dbus_connection(bus="SYSTEM")

    def get_info(self):
        """Gather information from the system"""
        info = {}
        for service in self.services:
            try:
                # Construct a message to get the 'ActiveState' property
                dbus_service = (
                    service.replace("-", "_2d")
                    .replace(".", "_2e")
                    .replace("\\", "_5c")
                    .replace("@", "_40")
                )

                msg = new_method_call(
                    DBusAddress(
                        object_path=f"/org/freedesktop/systemd1/unit/{dbus_service}",
                        bus_name="org.freedesktop.systemd1",
                        interface="org.freedesktop.DBus.Properties",
                    ),
                    "Get",
                    "ss",
                    ("org.freedesktop.systemd1.Unit", "ActiveState"),
                )

                # Send the message and get the reply
                reply = self.bus.send_and_get_reply(msg)

                # The actual state is wrapped in a variant 'v'
                state = reply.body[0][1]
                status = "ON" if state == "active" else "OFF"
            except Exception as e:
                logger.error("Error getting status for %s: %s", service, e)
                status = "OFF"

            name = service.replace(".service", "")
            info[name] = status
        return info

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for service in self.services:
            name = service.replace(".service", "")
            discovery_info[f"Systemd {name}"] = {
                "type": "switch",
                "icon": "mdi:application-cog",
                "value_template": f"{{{{ value_json.get('{name}') }}}}",
            }
        return discovery_info

    def start_control(self, topic, data):
        """Control system"""
        service_name = topic[1].replace("systemd_", "") + ".service"
        action = "StartUnit" if data.lower() == "on" else "StopUnit"

        try:
            # Create a new method call message
            msg = new_method_call(
                DBusAddress(
                    object_path="/org/freedesktop/systemd1",
                    bus_name="org.freedesktop.systemd1",
                    interface="org.freedesktop.systemd1.Manager",
                ),
                action,
                "ss",  # signature for two strings
                (service_name, "replace"),
            )
            # Send the message
            self.bus.send(msg)
        except Exception as err:
            logger.error("Error %sing %s: %s", action, service_name, err)
            fallback_command = "start" if data.lower() == "on" else "stop"
            syscommand(f"sudo -n systemctl {fallback_command} {service_name} &")

    def __del__(self):
        """Close the connection"""
        if hasattr(self, "bus"):
            self.bus.close()
